"""
Custom Portfolio optimization environments
"""
import pandas as pd
import numpy as np
import torch
from .base import PortfolioOptimizationEnv
from .metrics import sharpe_ratio, tracking_error
from .common import create_return_matrices, decompose_weights_tensor

class SharpeEnv(PortfolioOptimizationEnv):
    """
    This class implements a custom environment following the `gym` structure for Portfolio Optimization.
    Using the Sharpe ratio as the rewards

    :param pd.DataFrame df_ohlc: pd.DataFrame with the OHLC+ of the portfolio instruments. Columns are multiindex (Symbol, Price) e.g. ('AAPL', 'Close')
    :param pd.DataFrame df_observations: pd.DataFrame with the environment features.
    :param int rebalance_every: periods between consecutive rebalancing actions.
    :param float slippage: %loss due to gap between decision price for the agent and the execution price.
    :param float transaction_costs: %loss due to execution of the trade.
    :param bool continuous_weights: `True` to split the weights in (h)eld, (b)ought and (s)old positions.
    :param bool allow_short_positions: `True` to enable short positions.
    :param int max_trajectory_len: max total number of periods for the trajectories. E.g. 252 for a trading year.
    :param int observation_frame_lookback: return the previous N observations from the environment to take the next action.
    :param str render_mode: Either `tile` (2D), `tensor`(+2D) or `vector` (1D) to return the environment state.
    :param str agent_type: `discrete` or `continuous`.
    :param bool convert_to_terminated_truncated: use done (old Gym version) or truncated and terminated (new Gymnasium version)
    :param float riskfree_rate: risk free rate to compute the numerator of the sharpe ratio (r-Rf)
    :param int periods_per_year: periods per year to annualize returns for the Sharpe ratio computation.
    :param bool compute_cumulative: use the whole trajectory of the episode or just the last returns for the computation
    :param verbose: verbosity (0: None, 1: error messages, 2: all messages)
    """

    def __init__(self,
                 df_ohlc: pd.DataFrame,
                 df_observations: pd.DataFrame,
                 rebalance_every: int = 1,
                 slippage: float = 0.0005,
                 transaction_costs: float = 0.0002,
                 continuous_weights: bool = False,
                 allow_short_positions: bool = False,
                 max_trajectory_len: int = 252,
                 observation_frame_lookback: int = 5,
                 render_mode: str = 'tile',
                 agent_type: str = 'discrete',
                 convert_to_terminated_truncated: bool = False,
                 riskfree_rate: float = 0.0,
                 periods_per_year:int = 252,
                 compute_cumulative:bool = False,
                 verbose: int = 0,
                 ):

        super().__init__(df_ohlc, df_observations, rebalance_every, slippage, transaction_costs,
                         continuous_weights, allow_short_positions, max_trajectory_len, observation_frame_lookback,
                         render_mode, agent_type, convert_to_terminated_truncated, verbose)

        self.riskfree_rate = riskfree_rate
        self.periods_per_year = periods_per_year
        self.compute_cumulative = compute_cumulative

    def compute_reward(self, r) -> torch.Tensor:
        """
        :param r: returns series
        :return: sharpe ratio of the returns or the full trajectory
        """
        return sharpe_ratio(r if not self.compute_cumulative else self.trajectory_returns,
                            self.riskfree_rate, self.periods_per_year)



class TrackingErrorEnv(PortfolioOptimizationEnv):
    """
     This class implements a custom environment following the `gym` structure for Portfolio Optimization.
     Using the Tracking error as the reward

    :param pd.DataFrame df_ohlc: pd.DataFrame with the OHLC+ of the portfolio instruments. Columns are multiindex (Symbol, Price) e.g. ('AAPL', 'Close')
    :param pd.DataFrame df_observations: pd.DataFrame with the environment features.
    :param pd.DataFrame df_reference: pd.DataFrame with the reference returns (tracked instrument).
    :param int rebalance_every: periods between consecutive rebalancing actions.
    :param float slippage: %loss due to gap between decision price for the agent and the execution price.
    :param float transaction_costs: %loss due to execution of the trade.
    :param bool continuous_weights: `True` to split the weights in (h)eld, (b)ought and (s)old positions.
    :param bool allow_short_positions: `True` to enable short positions.
    :param int max_trajectory_len: max total number of periods for the trajectories. E.g. 252 for a trading year.
    :param int observation_frame_lookback: return the previous N observations from the environment to take the next action.
    :param str render_mode: Either `tile` (2D), `tensor`(+2D) or `vector` (1D) to return the environment state.
    :param str agent_type: `discrete` or `continuous`.
    :param bool convert_to_terminated_truncated: use done (old Gym version) or truncated and terminated (new Gymnasium version)
    :param verbose: verbosity (0: None, 1: error messages, 2: all messages)
     """

    def __init__(self,
                 df_ohlc: pd.DataFrame,
                 df_observations: pd.DataFrame,
                 df_reference: pd.DataFrame,
                 rebalance_every: int = 1,
                 slippage: float = 0.005,
                 transaction_costs: float = 0.002,
                 continuous_weights: bool = False,
                 allow_short_positions: bool = False,
                 max_trajectory_len: int = 252,
                 observation_frame_lookback: int = 5,
                 render_mode: str = 'tile',
                 agent_type: str = 'discrete',
                 convert_to_terminated_truncated: bool = False,
                 verbose: int = 0
                ):

        self.df_reference = df_reference
        self.reference_returns = None
        self.trajectory_reference_returns = None
        super().__init__(df_ohlc, df_observations, rebalance_every, slippage, transaction_costs,
                         continuous_weights, allow_short_positions, max_trajectory_len, observation_frame_lookback,
                         render_mode, agent_type, convert_to_terminated_truncated, verbose)

    def reset(self, seed: int = 5106, options: dict = None) -> tuple:
        """
        Method to reset the environment, it empties the trajectory reference returns.
        :param int seed: random seed.
        :param dict options: dictionary of options
        :return:
        """
        self.trajectory_reference_returns = []
        obs, info = super().reset()
        return obs, info

    def compute_reward(self, r) -> torch.Tensor:
        """
        Computes the tracking error. It clips the values for stability and helping convergence.
        :param pd.Series r: returns series
        :return: tracking error of the episode trajectory VS the reference trajectory.
        """
        if isinstance(self.trajectory_returns, list):
            trajectory_rets = pd.concat(self.trajectory_returns)
        else:
            trajectory_rets = self.trajectory_returns

        trajectory_rets = torch.Tensor(trajectory_rets).squeeze()

        reference_rets = pd.concat(self.trajectory_reference_returns).drop_duplicates()
        reference_rets = reference_rets.tail(reference_rets.shape[0]-1)
        reference_rets = torch.Tensor(reference_rets).squeeze()

        terror = tracking_error(trajectory_rets, reference_rets)
        creward = torch.clip(torch.pow(terror, -1), 0, 1e5)
        return creward

    def step(self, action) -> tuple:
        """
        Environment step method. Takes the agent's action and returns the new state, reward and whether or not the episode has finished.
        This method allows legacy behavior of gym when an episode has finished, returning the `done` boolean flag, and the new
        gymnasium convention of separating the `done` flag into `terminated` (the episode reached the max length allowed) or
        `truncated` if it ended for different reasons.

        The step method does the following sequence of operations in the base environment

        1. Compute the new weights based on the action taken.
        Depending on the agent, it can be either a full allocation to an instrument (`discrete`), 1-sum weights (`continuous`
        action space with long-only positions) or 0-sum weights (`continuous` action space with dollar-neutral positions).

        2. Compute the returns of the held, bought and sold positions.
        - Held positions return are computed as the price at last closing period divided by the initial open period.
        - Bought positions are bought at the Open. Return is the close by the open prices of that period.
        - Sold positions are sold at the Open. Return is the price at the open divided by the previous period close.

        3. Get the observation frame and expand if necessary. Transform into vector (1D), tile (2D) or tensor (3+D).

        4. Check if the episode has ended (truncated / terminated) and update the date tracker.

        5. Split the weights in held, bought and sold positions to compute the return series and append to the returns dataframe.

        6. Return the new state, reward, done flag (or truncated and terminated) and information.

        Then, in the outer part of the `step()` method:

        1. The return tuple from the base environment is unrolled.

        2. The tracking error reward is computed.

        3. New boolean flags for done  / terminated and truncated are computed.

        :param int or float action: index of the action for `discrete` agents or float values for each dimension of the action space for continuous agents.
        :return: the new state, reward, done flag (or truncated and terminated) and information.
        :rtype: tuple
        """
        self.reference_returns = self.df_reference[self.current_rebalancing_date:self.next_rebalancing_date]
        self.trajectory_reference_returns.append(self.reference_returns)

        _ = super().step(action)

        if not self.convert_to_terminated_truncated:
            observations, reward, done, info = _
        else:
            observations, reward, truncated, terminated, info = _

        reward = self.compute_reward(self.reference_returns)

        if self.convert_to_terminated_truncated:
            return observations, reward, truncated, terminated, info
        else:
            return observations, reward, done, info



def reward_criterion_scorer(func, r_a, r_b, threshold, **kwargs):
    """
    This function computes the normalized score of a reward criterion to compare 2 sets of portfolio returns
    :param func: scoring function
    :param r_a: returns of the first strategy
    :param r_b: returns of the second strategy
    :return:
    """
    sc_a = func(r_a, **kwargs)
    sc_b = func(r_b, **kwargs)

    ratio = sc_a/sc_b

    if np.abs(ratio-1) <= threshold:
        return 0.5
    elif ratio-1 > 0:
        return 1
    else:
        return 0

class CompositeRewardEnv(PortfolioOptimizationEnv):
    """
     This class implements a custom environment following the `gym` structure for Portfolio Optimization.
     Using a composite reward mechanism and pseudo-inverse rl approach.

     The composite reward is a set of functions that take the returns of two different strategies and provide a score
     across the individual criteria and aggregate them to determine if there is an overall winner (reward=0 or 1)
     or if both are comparable (0.5)

    :param pd.DataFrame df_ohlc: pd.DataFrame with the OHLC+ of the portfolio instruments. Columns are multiindex (Symbol, Price) e.g. ('AAPL', 'Close')
    :param pd.DataFrame df_observations: pd.DataFrame with the environment features.
    :param list composite_reward_funcs: list of criteria functions to compute the composite reward with.
    :param float composite_reward_threshold: threshold to consider comparable results (0.5 +/- threshold).
    :param int rebalance_every: periods between consecutive rebalancing actions.
    :param float slippage: %loss due to gap between decision price for the agent and the execution price.
    :param float transaction_costs: %loss due to execution of the trade.
    :param bool continuous_weights: `True` to split the weights in (h)eld, (b)ought and (s)old positions.
    :param bool allow_short_positions: `True` to enable short positions.
    :param int max_trajectory_len: max total number of periods for the trajectories. E.g. 252 for a trading year.
    :param int observation_frame_lookback: return the previous N observations from the environment to take the next action.
    :param str render_mode: Either `tile` (2D), `tensor`(+2D) or `vector` (1D) to return the environment state.
    :param str agent_type: `discrete` or `continuous`
    :param bool convert_to_terminated_truncated: use done (old Gym version) or truncated and terminated (new Gymnasium version)
    :param verbose: verbosity (0: None, 1: error messages, 2: all messages)
     """

    def __init__(self,
                 df_ohlc: pd.DataFrame,
                 df_observations: pd.DataFrame,
                 composite_reward_funcs: list,
                 composite_reward_threshold: float = 0.1,
                 rebalance_every: int = 1,
                 slippage: float = 0.005,
                 transaction_costs: float = 0.002,
                 continuous_weights: bool = False,
                 allow_short_positions: bool = False,
                 max_trajectory_len: int = 252,
                 observation_frame_lookback: int = 5,
                 render_mode: str = 'tile',
                 agent_type: str = 'discrete',
                 convert_to_terminated_truncated: bool = False,
                 verbose: int = 0,
                 **kwargs
                 ):

        self.kwargs = kwargs
        self.prior_weights = None
        self.composite_reward_funcs = composite_reward_funcs
        self.composite_reward_threshold = composite_reward_threshold
        super().__init__(df_ohlc, df_observations, rebalance_every, slippage, transaction_costs,
                         continuous_weights, allow_short_positions, max_trajectory_len, observation_frame_lookback,
                         render_mode, agent_type, convert_to_terminated_truncated, verbose)

    def reset(self, seed: int = 5106, options: dict = None) -> tuple:
        """
        Method to reset the environment, it empties the trajectory reference returns.
        :param int seed: random seed.
        :param dict options: dictionary of options
        :return:
        """
        obs, info = super().reset()
        return obs, info

    def compute_composite_rewards(self, r_new, r_old) -> torch.Tensor:
        """
        Computes the composite reward.
        Each function takes the returns with the new and former portfolio weights
        :param pd.Series r: returns series
        :return: tracking error of the episode trajectory VS the reference trajectory.
        """
        c = 0
        for func in self.composite_reward_funcs:
            score = reward_criterion_scorer(func,  r_new, r_old,
                                            self.composite_reward_threshold, **self.kwargs)
            c += score

        c /= len(self.composite_reward_funcs)

        if c >= 0.5+self.composite_reward_threshold:
            return 1
        elif c < 0.5-self.composite_reward_threshold:
            return 0
        else:
            return 0.5


    def step(self, action) -> tuple:
        """
        Environment step method. Takes the agent's action and returns the new state, reward and whether or not the episode has finished.
        This method allows legacy behavior of gym when an episode has finished, returning the `done` boolean flag, and the new
        gymnasium convention of separating the `done` flag into `terminated` (the episode reached the max length allowed) or
        `truncated` if it ended for different reasons.

        The step method does the following sequence of operations in the base environment

        1. Compute the new weights based on the action taken.
        Depending on the agent, it can be either a full allocation to an instrument (`discrete`), 1-sum weights (`continuous`
        action space with long-only positions) or 0-sum weights (`continuous` action space with dollar-neutral positions).

        2. Compute the returns of the held, bought and sold positions.
        - Held positions return are computed as the price at last closing period divided by the initial open period.
        - Bought positions are bought at the Open. Return is the close by the open prices of that period.
        - Sold positions are sold at the Open. Return is the price at the open divided by the previous period close.

        3. Get the observation frame and expand if necessary. Transform into vector (1D), tile (2D) or tensor (3+D).

        4. Check if the episode has ended (truncated / terminated) and update the date tracker.

        5. Split the weights in held, bought and sold positions to compute the return series and append to the returns dataframe.

        6. Return the new state, reward, done flag (or truncated and terminated) and information.

        Then, in the outer part of the `step()` method:

        1. The return tuple from the base environment is unrolled.

        2. The tracking error reward is computed.

        3. New boolean flags for done  / terminated and truncated are computed.

        :param int or float action: index of the action for `discrete` agents or float values for each dimension of the action space for continuous agents.
        :return: the new state, reward, done flag (or truncated and terminated) and information.
        :rtype: tuple
        """

        # Check that weights have the correct dimension ---

        # assert action.shape[1] == self.action_size
        self.current_weights = self.new_weights

        if self.agent_type == 'discrete':
            self.new_weights = torch.zeros(self.action_size)
            self.new_weights[action] = 1.0
        else:
            if not self.allow_short_positions:
                self.new_weights = torch.Tensor(action) / torch.Tensor(action).sum()
            else:
                self.new_weights = torch.Tensor(action) - torch.Tensor(action).mean()

                # Get the observation frame ---
        """
        Observation frame takes from the Action date (consecutive timestamp from previous decision date where buys/sell became effective, 
        to the decision date at close)
        """
        # get trajectory returns during holding period ---
        effective_rebalancing_date = self.available_dates[self.available_dates.index(self.current_rebalancing_date) + 1]
        r_sell = torch.Tensor(self.returns_sell.loc[[effective_rebalancing_date]].values).squeeze()
        r_buy = torch.Tensor(self.returns_buy.loc[[effective_rebalancing_date]].values).squeeze()
        return_frame = self.returns_hold.loc[effective_rebalancing_date:self.next_rebalancing_date, :]
        R_hold = torch.Tensor(return_frame.values)
        r_hold = torch.Tensor(self.returns_hold.loc[[effective_rebalancing_date]].values).squeeze()

        # Observation frame is the next information available, from our action date (decision date+1) to the next rebalancing date
        # This is the information that we will use to decide the next weights.
        idx_lookback = max(0, self.available_dates.index(self.next_rebalancing_date) - self.observation_frame_lookback)

        observation_frame = self.df_observations[self.available_dates[idx_lookback]:self.next_rebalancing_date]
        observation_frame = self.expand_observation_frame(observation_frame)

        # Add any additional information ---
        info = dict()
        info['indices'] = observation_frame.index.tolist()
        info['features'] = observation_frame.columns.tolist()

        if self.render_mode == 'tile':
            observations = self.return_obs_frame_as_tile(observation_frame)
        elif self.render_mode == 'vector':
            observations = self.return_obs_frame_as_vector(observation_frame)
        elif self.render_mode == 'tensor':
            observations = self.return_obs_frame_as_tensor(observation_frame)

        # Check if the environment is done ---
        """
        If the next rebalancing date is the last one, we will not be able to compute the rewards afterwards. 
        """
        self.current_trajectory_len += self.rebalance_every
        truncated = False
        terminated = False
        if (self.next_rebalancing_date == self.rebalancing_dates[-1]):
            # We do not rebalance in the last rebalancing date, we just want to keep homogeneous decision tensors.
            terminated = True
        elif (self.current_trajectory_len == self.max_trajectory_len):
            truncated = True
        else:
            # move the date cursor to the next rebalancing dates ---
            self.current_rebalancing_date = self.next_rebalancing_date  # Date when we are taking the decision
            self.next_rebalancing_date = self.rebalancing_dates[
                self.rebalancing_dates.index(self.current_rebalancing_date) + 1]  # Get next rebalancing date

        # Compute split weights and compute returns ---
        if self.continuous_weights:
            r = torch.matmul(self.new_weights, R_hold.T)
            r[0] += - (self.transaction_costs + self.slippage)

            r_o = torch.matmul(self.current_weights, R_hold.T)
            r_o[0] += - (self.transaction_costs + self.slippage)
        else:
            w_h, w_b, w_s = decompose_weights_tensor(self.new_weights, self.current_weights)
            # Compute returns of Buy/Sell
            r_s = torch.dot(w_s.squeeze(), r_sell)
            r_b = torch.dot(w_b.squeeze(), r_buy)
            r_h = torch.dot(w_h.squeeze(), r_hold)

            # Multiply returns x weights to get the portfolio returns
            r = torch.matmul(self.new_weights, R_hold.T)

            # Modify the returns of the 1st day of trajectory to account for slippage and transaction costs
            # as well as buys/sales returns
            r[0] += (r_s - self.transaction_costs - self.slippage)
            r[0] += (r_b - self.transaction_costs - self.slippage)
            r[0] += r_h

            # Compute returns with previous weights
            r_o = torch.matmul(self.current_weights, R_hold.T)

        self.last_returns = r
        self.prior_returns = r_o

        df_r = pd.Series(r.numpy().squeeze(), index=pd.to_datetime(return_frame.index))
        self.trajectory_returns.append(df_r)
        if truncated or terminated:
            self.trajectory_returns = pd.concat(self.trajectory_returns)

        reward = self.compute_composite_rewards(r, r_o)

        if self.convert_to_terminated_truncated:
            return observations, reward, truncated, terminated, info
        else:
            return observations, reward, truncated or terminated, info


import numpy as np
import pandas as pd
import torch


def get_price_window(df: pd.DataFrame, ref_date: str, lookback_window: int) -> pd.DataFrame:
    """
    Creates a price window given a history dataframe, using the last reference date and a lookback window.
    :param df: dataframe of prices
    :param ref_date: date of the last available prices
    :param int lookback_window: periods to look back in history to get the window. Negative values to take all available history.
    :return: Last lookback_window observations of a table.
    :rtype: pd.DataFrame
    """
    _ = df[df.index <= ref_date].sort_index()
    return _ if lookback_window < 0 else _.tail(lookback_window)


def decompose_weights_df(df_weights: pd.DataFrame) -> tuple:
    """
    This function takes a dataframe of portfolio weights and decomposes in 3 matrices:
    W_h, W_s, W_b. Satisfying this equation:
    W_h(t) = W_n where W_n==W_o
    W_s = W_h(t)-W_h(t-1) <0 : negative weight delta, necessary sales
    W_b = W_h(t)-W_h(t-1) >0 : positive weight delta, necessary buys

    :param pd.DataFrame df_weights: weights dataframe
    :return: The weights of the instruments to hold, buy or sale in that order
    :rtype: tuple
    """
    deltas = df_weights.diff()
    w_b = np.abs(deltas * (deltas >= 0))
    w_s = deltas * (deltas <= 0)
    w_h = df_weights * (df_weights <= df_weights.shift(1)) + df_weights.shift(1) * (df_weights > df_weights.shift(1))
    return w_h, w_b, w_s


def decompose_weights_tensor(w_n: torch.Tensor, w_o: torch.Tensor) -> tuple:
    """
    This function takes the old and new tensor of weights in the portfolio and decomposes
    it in the (h)eld, (b)ought and (sold) assets W_h, W_s, W_b. Satisfying this equation:
    W_h(t) = W_n where W_n==W_o
    W_s = W_h(t)-W_h(t-1) <0 : negative weight delta, necessary sales
    W_b = W_h(t)-W_h(t-1) >0 : positive weight delta, necessary buys

    :param w_n: new weight positions
    :param w_o: previous weight positions
    :return: w_h (Weights of instruments to hold), w_b (Necessary buys), w_s (Necessary sales)
    :rtype: tuple
    """
    w_delta = w_n - w_o
    w_b = torch.abs(w_delta * (w_delta > 0)).float()
    w_s = torch.abs(w_delta * (w_delta < 0)).float()
    w_h = 0.5*(w_o+w_n-w_s-w_b).float()
    assert torch.max(torch.abs(w_h+w_b-w_n))<1e-6
    return w_h, w_b, w_s


def compute_OHLC_returns(df_ohlc: pd.DataFrame) -> pd.DataFrame:
    """
    This function takes an OHLC matrix of prices and computes the daily returns for each of the available actions.
    For daily returns:
    (Buy) action has a daily return of Close_t/Open_t
    (Sell) action has a daily return of the overnight change Open_t/Close_t-1
    (Hold) action has a daily return of Close_t/Close_t-1

    For other periodicitie:
    (Buy) action has a period-over-period return of Close_t/Open_t, assuming that it is immediately bought at the beginning of period t.
    (Sell) action has a period-over-period return Open_t/Close_t-1, assuming that it is immediately sold at the beginning of period t.
    (Hold) action has a period-over-period return of Close_t/Close_t-1.

    :param pd.DataFrame df_ohlc: OHLC returns of the assets
    :return: Dataframe with the returns of the held/bought/sold assets.
    :rtype: pd.DataFrame
    """
    df_rets = df_ohlc.copy(deep = True)
    df_rets['r_b'] = (df_rets['Close'] / df_rets['Open']) - 1
    df_rets['r_s'] = (df_rets['Open'] / df_rets['Close'].shift(1)) - 1
    df_rets['r_h'] = (df_rets['Close'] / df_rets['Close'].shift(1)) - 1

    return df_rets.loc[:, ['r_h', 'r_b', 'r_s']]


def create_return_matrices(df_instruments: pd.DataFrame) -> tuple:
    """
    This function takes a dataframe of OHLC instruments and computes the return matrices Rh, Rb and Rs for each action
    (Buy) action has a daily return of Close_t/Open_t
    (Sell) action has a daily return of the overnight change Open_t/Close_t-1
    (Hold) action has a daily return of Close_t/Close_t-1
    :param df_instruments: dataframe with OHLC of the data and column index [i,j] where i is the instrument and j the OHLC prices
    :return: typle with the return dataframes of held, bought and sold assets
    :rtype: tuple
    """
    r_h = dict()
    r_b = dict()
    r_s = dict()
    for i in set([i[0] for i in df_instruments.columns]):
        _ = compute_OHLC_returns(df_instruments[i])
        r_h[i] = _['r_h'].values
        r_b[i] = _['r_b'].values
        r_s[i] = _['r_s'].values

    r_h = pd.DataFrame(r_h, index=df_instruments.index).dropna()
    r_b = pd.DataFrame(r_b, index=df_instruments.index).dropna()
    r_s = pd.DataFrame(r_s, index=df_instruments.index).dropna()

    return r_h, r_b, r_s


def normalize_price_window(df_price_window: pd.DataFrame, norm_pivot: str = 'window_close') -> pd.DataFrame:
    """
    This function normalizes an OHLC dataframe of N days based on a criterion. window_close uses the last closing price,
    window_open the first open price. prev_day_close / day_open the first or last price available for the day.
    Normalization is done as P/normalization reference.
    :param df_price_window: OHLC prices
    :param norm_pivot: name of the normalization strategy. One of window_close(open), day_close(open).
    :return: normalized df_price_window.
    :rtype: pd.DataFrame
    """
    if norm_pivot == 'prev_day_close':
        df_norm_price = df_price_window.div(df_price_window['Close'].shift(1), axis=0)
    elif norm_pivot == 'day_close':
        df_norm_price = df_price_window.div(df_price_window['Close'], axis=0)
    elif norm_pivot == 'day_open':
        df_norm_price = df_price_window.div(df_price_window['Open'], axis=0)
    elif norm_pivot == 'window_close':
        df_norm_price = df_price_window / df_price_window['Close'][-1]
    elif norm_pivot == 'window_open':
        df_norm_price = df_price_window / df_price_window['Open'][0]
    else:
        raise ValueError(f"{norm_pivot} is not a valid normalization strategy")

    return df_norm_price






import numpy as np
import pandas as pd
import random
import torch
from .common import create_return_matrices, decompose_weights_tensor
import gymnasium
from gymnasium.spaces import Box, Discrete
import collections


class PortfolioOptimizationEnv(gymnasium.Env):
    """
    This class implements a custom environment following the `gym` structure for Portfolio Optimization.
    The vanilla environment uses the average return as the reward
    """

    def __init__(self,
                 df_ohlc: pd.DataFrame,
                 df_observations: pd.DataFrame,
                 rebalance_every: int = 1,
                 slippage: float = 0.0005,
                 transaction_costs: float = 0.0002,
                 continuous_weights: bool = False,
                 allow_short_positions: bool = False,
                 max_trajectory_len: int = 252,
                 observation_frame_lookback: int = 5,
                 render_mode: str = 'tile',
                 agent_type: str = 'discrete',
                 convert_to_terminated_truncated: bool = False,
                 trajectory_bootstrapping: bool = False,
                 epìsodic_instrument_shifting: bool = False,
                 verbose: int = 0,
                 ):
        """
        :param pd.DataFrame df_ohlc: pd.DataFrame with the OHLC+ of the portfolio instruments. Columns are multiindex (Symbol, Price) e.g. ('AAPL', 'Close')
        :param pd.DataFrame df_observations: pd.DataFrame with the environment features.
        :param int rebalance_every: periods between consecutive rebalancing actions.
        :param float slippage: %loss due to gap between decision price for the agent and the execution price.
        :param float transaction_costs: %loss due to execution of the trade.
        :param bool continuous_weights: `False` to split the weights in (h)eld, (b)ought and (s)old positions.
        :param bool allow_short_positions: `True` to enable short positions.
        :param int max_trajectory_len: max total number of periods for the trajectories. E.g. 252 for a trading year.
        :param int observation_frame_lookback: return the previous N observations from the environment to take the next action.
        :param str render_mode: Either `tile` (2D), `tensor`(3D) or `vector`(1D) to return the environment state.
        :param str agent_type: `discrete` or `continuous`
        :param bool convert_to_terminated_truncated: use done (old Gym version) or truncated and terminated (new Gymnasium version)
        :param bool trajectory_bootstrapping: use non-consecutive ordered rebalancing dates to break correlation (trajectory bootstrapping)
        :param bool epìsodic_instrument_shifting: shift order of instruments at env.reset()
        :param verbose: verbosity (0: None, 1: error messages, 2: all messages)
        """
        self.indicator_instrument_names = None
        self.indicator_names = None
        self.global_columns = None
        self.indicator_columns = None
        self.returns_sell = None
        self.returns_buy = None
        self.returns_hold = None
        self.df_ohlc = df_ohlc.dropna()
        self.n_instruments = self.df_ohlc.loc[:, [i for i in self.df_ohlc.columns if i[1] == 'Close']].shape[1]

        self.df_observations = df_observations
        self.process_indicator_types()
        self.preprocess_returns()

        self.rebalance_every = rebalance_every
        self.slippage = slippage
        self.transaction_costs = transaction_costs
        self.continuous_weights = continuous_weights
        self.allow_short_positions = allow_short_positions
        self.available_dates = self.df_ohlc.sort_index().index.values.tolist()
        self.rebalancing_dates = None
        self.current_rebalancing_date = None
        self.next_rebalancing_date = None
        self.render_mode = render_mode
        self.agent_type = agent_type
        self.convert_to_terminated_truncated = convert_to_terminated_truncated
        self.trajectory_bootstrapping = trajectory_bootstrapping
        self.episodic_instrument_shifting = epìsodic_instrument_shifting
        self.max_trajectory_len = max_trajectory_len
        self.observation_frame_lookback = observation_frame_lookback
        self.current_trajectory_len = None
        self.trajectory_returns = []
        self.last_returns = None

        self.action_space = self.get_action_space()
        self.action_size = self.action_space.n if isinstance(self.action_space, Discrete) else self.action_space.shape[
            0]
        self.observation_space = self.get_observation_space()

        self.current_weights = np.zeros(self.action_size)
        self.new_weights = np.zeros(self.action_size)

        self.reset()

        self.verbose = verbose

    """
    Environment auxiliary methods
    """

    def get_action_space(self):
        """
        Infer action space as the number of instruments to choose from. This method is used by StableBaselines3 to
        setup the network architecture of the agents.
        Instruments should be provided to the environment using a pd.DataFrame with MultiIndex columns [(Ticker, Open), (Ticker,...),(Ticker, Close)]

        :return: number of "actions" (instruments) and their bounds in the continuous action space case.
        """
        n_instruments = len(set([i[0] for i in self.df_ohlc.columns]))
        if self.agent_type == 'discrete':
            return Discrete(n_instruments)
        elif self.agent_type == 'continuous':
            return Box(-np.ones(n_instruments) * self.allow_short_positions, np.ones(n_instruments))

    def get_observation_space(self):
        """
        Get the observation space for the agents. This method is used by StableBaselines3 to
        setup the network architecture of the agents.
        :return: observation space with the bounds across each element of the state.

        """
        if self.render_mode == 'vector':
            lows = np.tile(self.df_observations.min(axis=0).values, (1 + self.observation_frame_lookback, 1)).squeeze()
            highs = np.tile(self.df_observations.max(axis=0).values, (1 + self.observation_frame_lookback, 1)).squeeze()
            return Box(low=lows, high=highs, shape=[1, (1 + self.observation_frame_lookback)*self.df_observations.shape[1]])
        if self.render_mode == 'tile':
            lows = np.tile(self.df_observations.min(axis=0).values, (1 + self.observation_frame_lookback, 1))
            highs = np.tile(self.df_observations.max(axis=0).values, (1 + self.observation_frame_lookback, 1))
            return Box(low=lows, high=highs, shape=[1 + self.observation_frame_lookback, self.df_observations.shape[1]])
        elif self.render_mode == 'tensor':
            new_shape = (len(self.indicator_names), len(self.indicator_instrument_names))
            lows = np.tile(np.reshape(self.df_observations.min(axis=0).values, new_shape),
                           (1 + self.observation_frame_lookback, 1, 1))
            lows = lows.transpose(2, 0, 1)
            highs = np.tile(np.reshape(self.df_observations.max(axis=0).values, new_shape),
                            (1 + self.observation_frame_lookback, 1, 1))
            highs = highs.transpose(2, 0, 1)
            return Box(low=lows, high=highs,
                       shape=[len(self.indicator_instrument_names), 1 + self.observation_frame_lookback,
                              len(self.indicator_names)])
        else:
            raise ValueError(f"render_mode={self.render_mode} not supported. Should be tile/tensor/vector")

    def preprocess_returns(self) -> None:
        """
        Transform OHLC prices into buy/hold/sell returns
        - Buy: Close_t/Open_t
        - Hold: Close_t/Close_t-1
        - Sell: Open_t/Close_t-1

        It transforms the df_ohlc attribute and stores in three dataframes the Hold/Buy/Sell returns per instrument.

        :return: None
        """
        ohlc_cols = list(set([i[1] for i in self.df_ohlc.columns]))
        instrument_ordering = [(i, j) for i in self.indicator_instrument_names for j in ohlc_cols]
        r_h, r_b, r_s = create_return_matrices(self.df_ohlc.loc[:,instrument_ordering])
        self.returns_hold = r_h
        self.returns_buy = r_b
        self.returns_sell = r_s

    def process_indicator_types(self) -> None:
        """
        This auxiliary method processes the df_observations dataframe to extract per-instrument indicators and global indicators.
        This method is used to ensure that tensor construction for convolutional feature extractors as per the SB3 conventions
        is correct

        :return: None
        """
        indicator_names = [i[1] for i in self.df_observations.columns]
        occurrences = collections.Counter(indicator_names)
        single = [k for k, v in occurrences.items() if v == 1]
        multiple = [k for k, v in occurrences.items() if v > 1]

        indicator_columns = [i for i in self.df_observations.columns if i[1] in multiple]
        self.global_columns = [i for i in self.df_observations.columns if i[1] in single]

        self.indicator_names = list(set([i[1] for i in indicator_columns]))
        self.indicator_instrument_names = list(set([i[0] for i in indicator_columns]))

        if self.episodic_instrument_shifting:
            random.shuffle(self.indicator_instrument_names)

        # Reorder indicators as [(Instrument A, indicator 1), (Instrument A, indicator 2) ... (Instrument M, indicator N)]
        # To be used with Convolutional feature extractors in the policy
        self.indicator_columns = [(i, j) for i in self.indicator_instrument_names for j in self.indicator_names]
        self.df_observations = self.df_observations.loc[:,self.indicator_columns]


    def compute_reward(self, r) -> torch.Tensor:
        """
        Compute the rewards as the sum of log-returns.
        :param torch.Tensor r: returns series
        :return: sum of log returns.
        :rtype: torch.Tensor of type float.
        """
        return torch.sum(torch.log(1 + r))

    def create_info(self):
        """
        Placeholder method to return additional environment info for custom environments
        :return: info dictionary with metadata.
        :rtype: dict
        """
        return dict()

    def reset(self, seed: int = 5106, options: dict = None) -> tuple:
        """
        Method to reset the environment, as per the gym(nasium) conventions. It takes a random date in the history
        and prepares the rebalancing dates based on the `rebalance_every` attribute.

        Weights are initialized randomly at the beginning of every episode, which has been proved to speed up training,
        compared to an initial 100% allocation to a single instrument (e.g. cash) or zero.


        :param seed: int, random seed.
        :param options: dictionary of options
        :return: initial observation for the episode and dictionary with information.
        :rtype: tuple
        """
        self.current_rebalancing_date = random.choice(self.available_dates[:-(self.rebalance_every + 1)])
        self.current_trajectory_len = 0.0
        self.trajectory_returns = []

        # This is required for episodic instrument shifting
        self.process_indicator_types()
        self.preprocess_returns()

        if self.trajectory_bootstrapping:
            n_samples = 1+self.max_trajectory_len//self.rebalance_every
            self.rebalancing_dates = np.random.choice(
                self.available_dates[self.available_dates.index(self.current_rebalancing_date):],
                n_samples)
            self.rebalancing_dates.sort()
        else:
            self.rebalancing_dates = self.available_dates[
                                     self.available_dates.index(self.current_rebalancing_date)::self.rebalance_every]

        self.next_rebalancing_date = self.rebalancing_dates[
            self.rebalancing_dates.index(self.current_rebalancing_date) + 1]
        _new_weights = torch.rand(self.action_size)
        self.new_weights = _new_weights / _new_weights.sum()  # We start without any assets

        idx_lookback = max(0,
                           self.available_dates.index(self.current_rebalancing_date) - self.observation_frame_lookback)

        observation_frame = self.df_observations[self.available_dates[idx_lookback]:self.current_rebalancing_date]
        observation_frame = self.expand_observation_frame(observation_frame)
        # Add any additional information ---
        info = dict()
        info['indices'] = observation_frame.index.tolist()
        info['features'] = observation_frame.columns.tolist()

        if self.render_mode == 'tile':
            observation_frame = self.return_obs_frame_as_tile(observation_frame)
        elif self.render_mode == 'vector':
            observation_frame = self.return_obs_frame_as_vector(observation_frame)
        elif self.render_mode == 'tensor':
            observation_frame = self.return_obs_frame_as_tensor(observation_frame)

        return observation_frame, info

    def expand_observation_frame(self, obs_frame):
        """
        Method to control the observations at points in time where `observation_frame_lookback+1` observations are not
        available. When there are no available observations, the method returns the last observation of the environment (df_observations)
        repeated `observation_frame_lookback+1` times.
        When the available observations are lower than `observation_frame_lookback+1`, it returns a random sample of the expected size.
        In any other case, the original observation frame is returned.

        :param pd.DataFrame obs_frame: observation frame
        :return: an observation frame with the expected `observation_frame_lookback+1` number of observations
        :rtype: pd.DataFrame
        """

        if obs_frame.shape[0] == 0:
            return self.df_observations.tail(1).sample(self.observation_frame_lookback + 1, replace=True)
        elif obs_frame.shape[0] < self.observation_frame_lookback + 1:
            return obs_frame.sample(self.observation_frame_lookback + 1, replace=True)
        else:
            return obs_frame

    def return_obs_frame_as_tensor(self, obs_frame) -> torch.Tensor:
        """
        This method takes the observation frame and returns it as a 3+D tensor, to be used for example with the CNN
        feature extractors of stable baselines3. It relies on the auxiliary method `process_indicator_types` to identify
        the global indicators and instrument specific indicators to shape the tensor. Global indicators (information available across instruments)
        are repeated `n_instruments` times and concatenated with the indicators of each instrument.

        :param pd.DataFrame obs_frame:
        :return: tensor with the merged observations.
        :rtype: torch.Tensor
        """
        # Take repeated indicators and reshape that
        # Tensors are in the shape of Channels x Height x Width -> instruments x lookback x indicators
        n_channels = len(self.indicator_instrument_names)
        global_tensor = torch.tile(torch.Tensor(obs_frame.loc[:, self.global_columns].values), [n_channels, 1, 1])

        indicators_tensor = torch.Tensor(
            [obs_frame[i].loc[:, self.indicator_names].values for i in self.indicator_instrument_names])

        if not 0 in global_tensor.size():
            return torch.concat([indicators_tensor, global_tensor], 1)
        else:
            return indicators_tensor

    def return_obs_frame_as_vector(self, obs_frame) -> torch.Tensor:
        """
        This method returns the observation frame as a 1D vector.
        :param pd.DataFrame obs_frame: environment observation frame with `lookback_window+1` rows.
        :return: 1D vector of the environment state.
        :rtype: torch.Tensor
        """
        obs = torch.Tensor(obs_frame.values.squeeze())
        return obs

    def return_obs_frame_as_tile(self, obs_frame) -> torch.Tensor:
        """
        This method returns the observation frame as a 2D vector.
        :param pd.DataFrame obs_frame: environment observation frame with `lookback_window+1` rows and `indicators` columns.
        :return: 2D matrix of the environment state.
        :rtype: torch.Tensor
        """
        obs = torch.Tensor(obs_frame.values)
        return obs

    def step(self, action) -> tuple:
        """
        Environment step method. Takes the agent's action and returns the new state, reward and whether or not the episode has finished.
        This method allows legacy behavior of gym when an episode has finished, returning the `done` boolean flag, and the new
        gymnasium convention of separating the `done` flag into `terminated` (the episode reached the max length allowed) or
        `truncated` if it ended for different reasons.

        The step method does the following sequence of operations.

        1. Compute the new weights based on the action taken.
        Depending on the agent, it can be either a full allocation to an instrument (`discrete`), 1-sum weights (`continuous`
        action space with long-only positions) or 0-sum weights (`continuous` action space with dollar-neutral positions).

        2. Compute the returns of the held, bought and sold positions.
        - Held positions return are computed as the price at last closing period divided by the initial open period.
        - Bought positions are bought at the Open. Return is the close by the open prices of that period.
        - Sold positions are sold at the Open. Return is the price at the open divided by the previous period close.

        3. Get the observation frame and expand if necessary. Transform into vector (1D), tile (2D) or tensor (3+D).

        4. Check if the episode has ended (truncated / terminated) and update the date tracker.

        5. Split the weights in held, bought and sold positions to compute the return series and append to the returns dataframe.

        6. Return the new state, reward, done flag (or truncated and terminated) and information.

        :param int or float action: index of the action for `discrete` agents or float values for each dimension of the action space for continuous agents.
        :return: the new state, reward, done flag (or truncated and terminated) and information.
        :rtype: tuple
        """
        # Check that weights have the correct dimension ---

        # assert action.shape[1] == self.action_size
        self.current_weights = self.new_weights

        if self.agent_type == 'discrete':
            self.new_weights = torch.zeros(self.action_size)
            self.new_weights[action] = 1.0
        else:
            if not self.allow_short_positions:
                self.new_weights = torch.Tensor(action) / torch.Tensor(action).sum()
            else:
                self.new_weights = torch.Tensor(action) - torch.Tensor(action).mean()

                # Get the observation frame ---
        """
        Observation frame takes from the Action date (consecutive timestamp from previous decision date where buys/sell became effective, 
        to the decision date at close)
        """
        # get trajectory returns during holding period ---
        effective_rebalancing_date = self.available_dates[self.available_dates.index(self.current_rebalancing_date) + 1]
        r_sell = torch.Tensor(self.returns_sell.loc[[effective_rebalancing_date]].values).squeeze()
        r_buy = torch.Tensor(self.returns_buy.loc[[effective_rebalancing_date]].values).squeeze()
        return_frame = self.returns_hold.loc[effective_rebalancing_date:self.next_rebalancing_date, :]
        R_hold = torch.Tensor(return_frame.values)
        r_hold = torch.Tensor(self.returns_hold.loc[[effective_rebalancing_date]].values).squeeze()

        # Observation frame is the next information available, from our action date (decision date+1) to the next rebalancing date
        # This is the information that we will use to decide the next weights.
        idx_lookback = max(0, self.available_dates.index(self.next_rebalancing_date) - self.observation_frame_lookback)

        observation_frame = self.df_observations[self.available_dates[idx_lookback]:self.next_rebalancing_date]
        observation_frame = self.expand_observation_frame(observation_frame)

        # Add any additional information ---
        info = dict()
        info['indices'] = observation_frame.index.tolist()
        info['features'] = observation_frame.columns.tolist()

        if self.render_mode == 'tile':
            observations = self.return_obs_frame_as_tile(observation_frame)
        elif self.render_mode == 'vector':
            observations = self.return_obs_frame_as_vector(observation_frame)
        elif self.render_mode == 'tensor':
            observations = self.return_obs_frame_as_tensor(observation_frame)

        # Check if the environment is done ---
        """
        If the next rebalancing date is the last one, we will not be able to compute the rewards afterwards. 
        """
        self.current_trajectory_len += self.rebalance_every
        truncated = False
        terminated = False
        if (self.next_rebalancing_date == self.rebalancing_dates[-1]):
            # We do not rebalance in the last rebalancing date, we just want to keep homogeneous decision tensors.
            terminated = True
        elif (self.current_trajectory_len == self.max_trajectory_len):
            truncated = True
        else:
            # move the date cursor to the next rebalancing dates ---
            self.current_rebalancing_date = self.next_rebalancing_date  # Date when we are taking the decision
            self.next_rebalancing_date = self.rebalancing_dates[
                self.rebalancing_dates.index(self.current_rebalancing_date) + 1]  # Get next rebalancing date

        # Compute split weights and compute returns ---
        if self.continuous_weights:
            r = torch.matmul(self.new_weights, R_hold.T)
            r[0] += - (self.transaction_costs + self.slippage)

        else:
            w_h, w_b, w_s = decompose_weights_tensor(self.new_weights, self.current_weights)
            # Compute returns of Buy/Sell
            r_s = torch.dot(w_s.squeeze(), r_sell)
            r_b = torch.dot(w_b.squeeze(), r_buy)
            r_h = torch.dot(w_h.squeeze(), r_hold)

            # Multiply returns x weights to get the portfolio returns
            r = torch.matmul(self.new_weights, R_hold.T)

            # Modify the returns of the 1st day of trajectory to account for slippage and transaction costs
            # as well as buys/sales returns
            r[0] += (r_s - self.transaction_costs - self.slippage)
            r[0] += (r_b - self.transaction_costs - self.slippage)
            r[0] += r_h

        self.last_returns = r

        df_r = pd.Series(r.numpy().squeeze(), index=pd.to_datetime(return_frame.index))
        self.trajectory_returns.append(df_r)
        if truncated or terminated:
            self.trajectory_returns = pd.concat(self.trajectory_returns)
        reward = self.compute_reward(r)

        if self.convert_to_terminated_truncated:
            return observations, reward, truncated, terminated, info
        else:
            return observations, reward, truncated or terminated, info



# METRICS
import torch

def annualize_rets(r: torch.Tensor, periods_per_year: int = 252) -> torch.Tensor:
    """
    Returns the annualized returns from a series of returns
    :param r: return series
    :param int periods_per_year: number of periods per year, 252 by default for daily returns
    :return: annualized returns
    :rtype: torch.Tensor
    """
    compounded_growth = (1 + r).prod()
    n_periods = r.shape[0]
    return compounded_growth ** (periods_per_year / n_periods) - 1


def annualize_vol(r: torch.Tensor, periods_per_year: int = 252) -> torch.Tensor:
    """
    Annualizes the volatility of a set of returns
    :param r: return series
    :param int periods_per_year: 252 by default for daily returns
    :return: annualized volatility of the returns
    :rtype: torch.Tensor
    """
    return torch.std(r, unbiased=True) * (periods_per_year ** 0.5)


def sharpe_ratio(r: torch.Tensor, riskfree_rate:float, periods_per_year:int = 252) -> torch.Tensor:
    """
    Computes the annualized Sharpe ratio of a set of returns
    :param r: return series
    :param riskfree_rate: risk free rate
    :param int periods_per_year: number of periods per year, 252 by default for daily returns
    :return: Sharpe ratio of the returns series
    :rtype: torch.Tensor
    """
    # convert the annual riskfree rate to per period
    rf_per_period = (1 + riskfree_rate) ** (1 / periods_per_year) - 1
    excess_ret = r - rf_per_period
    ann_ex_ret = annualize_rets(excess_ret, periods_per_year)
    ann_vol = annualize_vol(r, periods_per_year)
    return ann_ex_ret / ann_vol


def sortino_ratio(r: torch.Tensor, riskfree_rate: float, periods_per_year: int = 252) -> torch.Tensor:
    """
    Computes the annualized sortino ratio of a set of returns using the semideviation
    :param r: return series
    :param float riskfree_rate: risk free rate
    :param int periods_per_year: number of periods per year, 252 by default for daily returns
    :return: the Sortino ratio of the returns series
    :rtype: torch.Tensor
    """
    rf_per_period = (1 + riskfree_rate) ** (1 / periods_per_year) - 1
    excess_ret = r - rf_per_period
    ann_ex_ret = annualize_rets(excess_ret, periods_per_year)
    ann_vol = annualize_vol(torch.masked_select(r, r < 0), periods_per_year)
    return ann_ex_ret / ann_vol


def tracking_error(r, ref) -> torch.Tensor:
    """
    Returns the Tracking Error of a series of returns r and a reference series of returns ref
    :param r: return series
    :param ref: reference (tracked index) returns
    :return: Tracking error between the returns series and reference returns
    :rtype: torch.Tensor
    """
    t_e = torch.std(torch.sub(r, ref, alpha=1), unbiased = True)
    if t_e.isnan():
        return -torch.std(torch.sub(r, ref, alpha=1), unbiased = False)
    else:
        return t_e

def calmar_ratio(r, periods_per_year: int = 252) -> torch.Tensor:
    """
    Computes the Calmar ratio (annualized returns / maximum drawdown)
    :param r: return series
    :param int periods_per_year: number of periods per year, 252 by default for daily returns
    :return: Calmar ratio of the returns series.
    :rtype: torch.Tensor
    """
    cumulative_returns = torch.cumprod(1+r, 0)
    cumulative_max = torch.cummax(cumulative_returns, 0).values
    drawdowns = (cumulative_max - cumulative_returns) / cumulative_max
    max_dd = torch.max(drawdowns)

    return annualize_rets(r, periods_per_year) / max_dd

